/*************************************************************************************************
 * Command line utility to manage an index
 *                                                      Copyright (C) 2003-2004 Mikio Hirabayashi
 * This file is part of Estraier, a personal full-text search system.
 * Estraier is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 * Estraier is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with Estraier;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA.
 *************************************************************************************************/


#include "estcommon.h"

#define TEXTSUFS       ".txt,.asc"       /* suffix list of plain text */
#define HTMLSUFS       ".html,.htm"      /* suffix list of HTML */
#define MIMESUFS       ".eml,.mht"       /* suffix list of MIME */
#define SUFWILDCARD    "*"               /* wild card matching any file */
#define EXTINFILE      "extin.tmp"       /* name of a temporary file for input */
#define EXTOUTFILE     "extout.tmp"      /* name of a temporary file for output */
#define DLFILTERSUF    '@'               /* suffix of a dynamic linking function */


/* global variables */
const char *progname = NULL;             /* program name */
int sigterm = FALSE;                     /* flag for termination signal */
CBLIST *tsuflist = NULL;                 /* list of suffixes of plain text */
CBLIST *hsuflist = NULL;                 /* list of suffixes of HTML */
CBLIST *msuflist = NULL;                 /* list of suffixes of MIME */
CBLIST *xsuflist = NULL;                 /* list of suffixes for the external filter */
int xtext = FALSE;                       /* whether to use the outer filter of plain text */
int xmime = FALSE;                       /* whether to use the outer filter of MIME */
const char *ptcode = NULL;               /* encoding of file path as title */
const char *ftcode = NULL;               /* encoding of file name as title */


/* function prototypes */
int main(int argc, char **argv);
void setsignals(void);
void sigtermhandler(int num);
void usage(void);
int runregister(int argc, char **argv);
int runrelate(int argc, char **argv);
int runpurge(int argc, char **argv);
int runoptimize(int argc, char **argv);
int runinform(int argc, char **argv);
int runmerge(int argc, char **argv);
int runpree(int argc, char **argv);
void otcb(const char *fname, ODEUM *odeum, const char *msg);
char *readstdin(int *sp);
int bwimatchlist(const char *str, const CBLIST *keys);
char *fgetl(FILE *IN);
void pdperror(const char *name);
void printferror(const char *format, ...);
void printfinfo(const char *format, ...);
const char *datestr(int t);
char *filetouri(const char *file);
char *uritofile(const char *uri);
int proclist(const char *name, const char *lfile, int force, int wmax,
             const char *xtype, const char *xcmd, const char *enc);
int procdir(const char *name, const char *dir, int force, int wmax,
            const char *xtype, const char *xcmd, const char *enc);
int indexdir(ODEUM *odeum, VILLA *mtdb, const char *name, const char *dir, int force, int wmax,
             const char *xtype, const char *xcmd, const char *enc);
int indexfile(ODEUM *odeum, VILLA *mtdb, const char *name, const char *file, int force, int wmax,
              const char *xtype, const char *xcmd, const char *enc);
int ishot(VILLA *mtdb, const char *path, int mtime);
int procrelate(const char *name, const char *prefix, int force);
int procpurge(const char *name, const char *prefix, int force);
int procoptimize(const char *name);
int procinform(const char *name);
int procmerge(const char *name, const CBLIST *elems);
int procpree(const char *path, int hb, int mb,
             const char *xtype, const char *xcmd, const char *enc);
ODDOC *docext(const char *path, const char *ptr, int size, const char *enc,
              const char *xtype, const char *xcmd);
void docltitle(ODDOC *doc, const char *path);
void showdoc(ODDOC *doc);


/* main routine */
int main(int argc, char **argv){
  int rv;
  estputenv("LANG", ESTLOCALE);
  estputenv("LC_ALL", ESTLOCALE);
  cbstdiobin();
  progname = argv[0];
  sigterm = FALSE;
  setsignals();
  if(argc < 2) usage();
  odsetotcb(otcb);
  rv = 0;
  if(!strcmp(argv[1], "register")){
    rv = runregister(argc, argv);
  } else if(!strcmp(argv[1], "relate")){
    rv = runrelate(argc, argv);
  } else if(!strcmp(argv[1], "purge")){
    rv = runpurge(argc, argv);
  } else if(!strcmp(argv[1], "optimize")){
    rv = runoptimize(argc, argv);
  } else if(!strcmp(argv[1], "inform")){
    rv = runinform(argc, argv);
  } else if(!strcmp(argv[1], "merge")){
    rv = runmerge(argc, argv);
  } else if(!strcmp(argv[1], "pree")){
    rv = runpree(argc, argv);
  } else if(!strcmp(argv[1], "version") || !strcmp(argv[1], "--version")){
    printf("Estraier: a personal full-text search system version %s\n", _EST_VERSION);
    printf("Powered by QDBM version %s\n", dpversion);
    printf("Copyright (c) 2000-2003 Mikio Hirabayashi\n");
    rv = 0;
  } else {
    usage();
  }
  return rv;
}


/* set signal handlers */
void setsignals(void){
  signal(SIGHUP, sigtermhandler);
  signal(SIGINT, sigtermhandler);
  signal(SIGQUIT, sigtermhandler);
  signal(SIGPIPE, sigtermhandler);
  signal(SIGTERM, sigtermhandler);
}


/* handler of termination signal */
void sigtermhandler(int num){
  signal(num, SIG_DFL);
  sigterm = TRUE;
  printfinfo("the termination signal %d catched", num);
}


/* print the usage and exit */
void usage(void){
  fprintf(stderr, "%s: indexer of document files\n", progname);
  fprintf(stderr, "\n");
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "  %s register [-list file] [-force] [-wmax num]"
          " [-tsuf sufs] [-hsuf sufs] [-msuf sufs] [-xsuf sufs type cmd]"
          " [-xt] [-xm] [-enc code] [-pt code] [-ft code] name [dir]\n", progname);
  fprintf(stderr, "  %s relate [-force] name [prefix]\n", progname);
  fprintf(stderr, "  %s purge [-force] name [prefix]\n", progname);
  fprintf(stderr, "  %s optimize name\n", progname);
  fprintf(stderr, "  %s inform name\n", progname);
  fprintf(stderr, "  %s merge name elems...\n", progname);
  fprintf(stderr, "  %s pree [-h] [-m] [-x type cmd] [-xt] [-xm] [-enc code] [file]\n",
          progname);
  fprintf(stderr, "  %s version\n", progname);
  fprintf(stderr, "\n");
  exit(1);
}


/* parse arguments of register command */
int runregister(int argc, char **argv){
  char *name, *dir, *lfile, *tsuf, *hsuf, *msuf, *xsuf, *xtype, *xcmd, *enc, path[ESTPATHBUFSIZ];
  int i, wmax, force, plen, rv;
  name = NULL;
  dir = NULL;
  lfile = NULL;
  tsuf = NULL;
  hsuf = NULL;
  msuf = NULL;
  xsuf = NULL;
  xtype = NULL;
  xcmd = NULL;
  enc = NULL;
  force = FALSE;
  wmax = -1;
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-list")){
        if(++i >= argc) usage();
        lfile = argv[i];
      } else if(!strcmp(argv[i], "-force")){
        force = TRUE;
      } else if(!strcmp(argv[i], "-wmax")){
        if(++i >= argc) usage();
        wmax = atoi(argv[i]);
      } else if(!strcmp(argv[i], "-tsuf")){
        if(++i >= argc) usage();
        tsuf = argv[i];
      } else if(!strcmp(argv[i], "-hsuf")){
        if(++i >= argc) usage();
        hsuf = argv[i];
      } else if(!strcmp(argv[i], "-msuf")){
        if(++i >= argc) usage();
        msuf = argv[i];
      } else if(!strcmp(argv[i], "-xsuf")){
        if(++i >= argc) usage();
        xsuf = argv[i];
        if(++i >= argc) usage();
        xtype = argv[i];
        if(++i >= argc) usage();
        xcmd = argv[i];
      } else if(!strcmp(argv[i], "-xt")){
        xtext = TRUE;
      } else if(!strcmp(argv[i], "-xm")){
        xmime = TRUE;
      } else if(!strcmp(argv[i], "-enc")){
        if(++i >= argc) usage();
        enc = argv[i];
      } else if(!strcmp(argv[i], "-pt")){
        if(++i >= argc) usage();
        ptcode = argv[i];
      } else if(!strcmp(argv[i], "-ft")){
        if(++i >= argc) usage();
        ftcode = argv[i];
      } else {
        usage();
      }
    } else if(!name){
      name = argv[i];
    } else if(!dir){
      dir = argv[i];
    } else {
      usage();
    }
  }
  if(!name) usage();
  if(!dir) dir = ESTCDIRSTR;
  plen = sprintf(path, "%s", dir);
  if(plen > 1 && path[plen-1] == ESTPATHCHR) path[plen-1] = '\0';
  tsuflist = cbsplit(tsuf ? tsuf : TEXTSUFS, -1, ",");
  hsuflist = cbsplit(hsuf ? hsuf : HTMLSUFS, -1, ",");
  msuflist = cbsplit(msuf ? msuf : MIMESUFS, -1, ",");
  xsuflist = xsuf ? cbsplit(xsuf, -1, ",") : cblistopen();
  if(lfile){
    rv = proclist(name, lfile, force, wmax, xtype, xcmd, enc);
  } else {
    rv = procdir(name, path, force, wmax, xtype, xcmd, enc);
  }
  cblistclose(xsuflist);
  cblistclose(msuflist);
  cblistclose(hsuflist);
  cblistclose(tsuflist);
  return rv;
}


/* parse arguments of relate command */
int runrelate(int argc, char **argv){
  char *name, *prefix;
  int i, force, rv;
  name = NULL;
  prefix = NULL;
  force = FALSE;
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-force")){
        force = TRUE;
      } else {
        usage();
      }
    } else if(!name){
      name = argv[i];
    } else if(!prefix){
      prefix = argv[i];
    } else {
      usage();
    }
  }
  if(!name) usage();
  if(!prefix) prefix = "";
  rv = procrelate(name, prefix, force);
  return rv;
}


/* parse arguments of purge command */
int runpurge(int argc, char **argv){
  char *name, *prefix;
  int i, force, rv;
  name = NULL;
  prefix = NULL;
  force = FALSE;
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-force")){
        force = TRUE;
      } else {
        usage();
      }
    } else if(!name){
      name = argv[i];
    } else if(!prefix){
      prefix = argv[i];
    } else {
      usage();
    }
  }
  if(!name) usage();
  if(!prefix) prefix = "";
  rv = procpurge(name, prefix, force);
  return rv;
}


/* parse arguments of optimize command */
int runoptimize(int argc, char **argv){
  char *name;
  int i, rv;
  name = NULL;
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      usage();
    } else if(!name){
      name = argv[i];
    } else {
      usage();
    }
  }
  if(!name) usage();
  rv = procoptimize(name);
  return rv;
}


/* parse arguments of inform command */
int runinform(int argc, char **argv){
  char *name;
  int i, rv;
  name = NULL;
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      usage();
    } else if(!name){
      name = argv[i];
    } else {
      usage();
    }
  }
  if(!name) usage();
  rv = procinform(name);
  return rv;
}


/* parse arguments of merge command */
int runmerge(int argc, char **argv){
  char *name;
  CBLIST *elems;
  int i, rv;
  name = NULL;
  elems = cblistopen();
  for(i = 2; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      usage();
    } else if(!name){
      name = argv[i];
    } else {
      cblistpush(elems, argv[i], -1);
    }
  }
  if(!name) usage();
  if(cblistnum(elems) < 1){
    cblistclose(elems);
    usage();
  }
  rv = procmerge(name, elems);
  cblistclose(elems);
  return rv;
}


/* parse arguments of pree command */
int runpree(int argc, char **argv){
  char *file, *xtype, *xcmd, *enc;
  int i, hb, mb, rv;
  file = NULL;
  xtype = NULL;
  xcmd = NULL;
  enc = NULL;
  hb = FALSE;
  mb = FALSE;
  for(i = 2; i < argc; i++){
    if(!file && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-h")){
        hb = TRUE;
      } else if(!strcmp(argv[i], "-m")){
        mb = TRUE;
      } else if(!strcmp(argv[i], "-x")){
        if(++i >= argc) usage();
        xtype = argv[i];
        if(++i >= argc) usage();
        xcmd = argv[i];
      } else if(!strcmp(argv[i], "-xt")){
        xtext = TRUE;
      } else if(!strcmp(argv[i], "-xm")){
        xmime = TRUE;
      } else if(!strcmp(argv[i], "-enc")){
        if(++i >= argc) usage();
        enc = argv[i];
      } else if(!strcmp(argv[i], "-pt")){
        if(++i >= argc) usage();
        ptcode = argv[i];
      } else if(!strcmp(argv[i], "-ft")){
        if(++i >= argc) usage();
        ftcode = argv[i];
      } else {
        usage();
      }
    } else if(!file){
      file = argv[i];
    } else {
      usage();
    }
  }
  rv = procpree(file, hb, mb, xtype, xcmd, enc);
  return rv;
}


/* report the outturn */
void otcb(const char *fname, ODEUM *odeum, const char *msg){
  char *name;
  if(!(name = odname(odeum))) name = cbmemdup("error", -1);
  printf("%s: INFO: %s: %s: fsiz=%.0f dnum=%d wnum=%d bnum=%d\n",
         progname, name, msg, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  fflush(stdout);
  free(name);
}


/* read the standard input */
char *readstdin(int *sp){
  char *buf;
  int i, blen, c;
  blen = 256;
  buf = cbmalloc(blen);
  for(i = 0; (c = getchar()) != EOF; i++){
    if(i >= blen - 1) buf = cbrealloc(buf, blen *= 2);
    buf[i] = c;
  }
  buf[i] = '\0';
  *sp = i;
  return buf;
}


/* case insensitive backward matching with a list */
int bwimatchlist(const char *str, const CBLIST *keys){
  const char *kbuf;
  int i, ksiz;
  for(i = 0; i < cblistnum(keys); i++){
    kbuf = cblistval(keys, i, &ksiz);
    if(!strcmp(kbuf, SUFWILDCARD)) return TRUE;
    if(ksiz > 0 && cbstrbwimatch(str, kbuf)) return TRUE;
  }
  return FALSE;
}


/* read a line */
char *fgetl(FILE *IN){
  char *buf;
  int c, len, blen;
  buf = NULL;
  len = 0;
  blen = 256;
  while((c = fgetc(IN)) != EOF){
    if(blen <= len) blen *= 2;
    buf = cbrealloc(buf, blen + 1);
    if(c == '\n') c = '\0';
    buf[len++] = c;
    if(c == '\0') break;
  }
  if(!buf) return NULL;
  buf[len] = '\0';
  return buf;
}


/* print an error message */
void pdperror(const char *name){
  printf("%s: ERROR: %s: %s\n", progname, name, dperrmsg(dpecode));
  fflush(stdout);
}


/* print formatted error string and flush the buffer */
void printferror(const char *format, ...){
  va_list ap;
  va_start(ap, format);
  printf("%s: ERROR: ", progname);
  vprintf(format, ap);
  putchar('\n');
  fflush(stdout);
  va_end(ap);
}


/* print formatted information string and flush the buffer */
void printfinfo(const char *format, ...){
  va_list ap;
  va_start(ap, format);
  printf("%s: INFO: ", progname);
  vprintf(format, ap);
  putchar('\n');
  fflush(stdout);
  va_end(ap);
}


/* get static string of the date in RFC822 format */
const char *datestr(int t){
  static char buf[64], *wp;
  struct tm *tp, lt, gt;
  time_t tt;
  int lag;
  tt = (time_t)t;
  if(!(tp = gmtime(&tt))) return "Thu, 01 Jan 1970 00:00:00 GMT";
  gt = *tp;
  if(!(tp = localtime(&tt))) return "Thu, 01 Jan 1970 00:00:00 GMT";
  lt = *tp;
  lag = (lt.tm_hour * 60 + lt.tm_min) - (gt.tm_hour * 60 + gt.tm_min);
  if(lt.tm_year > gt.tm_year){
    lag += 24 * 60;
  } else if(lt.tm_year < gt.tm_year){
    lag -= 24 * 60;
  } else if(lt.tm_mon > gt.tm_mon){
    lag += 24 * 60;
  } else if(lt.tm_mon < gt.tm_mon){
    lag -= 24 * 60;
  } else if(lt.tm_mday > gt.tm_mday){
    lag += 24 * 60;
  } else if(lt.tm_mday < gt.tm_mday){
    lag -= 24 * 60;
  }
  wp = buf;
  wp += strftime(buf, sizeof(buf) - 1, "%a, %d %b %Y %H:%M:%S", &lt);
  sprintf(wp, " %+03d%02d", lag / 60, lag % 60);
  return buf;
}


/* make the URI from file path */
char *filetouri(const char *file){
  CBLIST *list;
  char str[ESTPATHBUFSIZ], *wp, *enc;
  const char *name;
  int i, nsiz;
  sprintf(str, "%c", ESTPATHCHR);
  list = cbsplit(file, -1, str);
  wp = str;
  for(i = 0; i < cblistnum(list); i++){
    if(i > 0) *(wp++) = '/';
    name = cblistval(list, i, &nsiz);
    enc = cburlencode(name, nsiz);
    wp += sprintf(wp, "%s", enc);
    free(enc);
  }
  cblistclose(list);
  *wp = '\0';
  return cbmemdup(str, -1);
}


/* make the file path from a URI */
char *uritofile(const char *uri){
  char *path;
  int i;
  path = cburldecode(uri, NULL);
  for(i = 0; path[i] != '\0'; i++){
    if(path[i] == '/') path[i] = ESTPATHCHR;
  }
  return path;
}


/* processing with finding files in a list file */
int proclist(const char *name, const char *lfile, int force, int wmax,
             const char *xtype, const char *xcmd, const char *enc){
  ODEUM *odeum;
  VILLA *mtdb;
  FILE *IN;
  char *line, path[ESTPATHBUFSIZ];
  int err, fatal;
  if(!strcmp(lfile, "-")){
    IN = stdin;
  } else {
    if(!(IN = fopen(lfile, "rb"))){
      printferror("%s: file cannot be opened", lfile);
      return 1;
    }
  }
  printfinfo("%s: registration started", name);
  if(!(odeum = odopen(name, OD_OWRITER | OD_OCREAT))){
    pdperror(name);
    if(IN != stdin) fclose(IN);
    return 1;
  }
  sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTMTDBNAME);
  if(!(mtdb = vlopen(path, VL_OWRITER | VL_OCREAT | VL_OZCOMP, VL_CMPLEX))){
    pdperror(name);
    odclose(odeum);
    if(IN != stdin) fclose(IN);
    return 1;
  }
  vlsettuning(mtdb, ESTMTDBLRM, ESTMTDBNIM, ESTMTDBLCN, ESTMTDBNCN);
  printfinfo("%s: database opened: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  err = FALSE;
  while((line = fgetl(IN)) != NULL){
    if(sigterm){
      printferror("aborting due to a termination signal");
      free(line);
      err = TRUE;
      break;
    }
    if(!indexfile(odeum, mtdb, name, line, force, wmax, xtype, xcmd, enc)) err = TRUE;
    free(line);
  }
  fatal = odfatalerror(odeum);
  printfinfo("%s: database closing: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!vlclose(mtdb)){
    pdperror(name);
    err = TRUE;
  }
  if(!odclose(odeum)){
    pdperror(name);
    err = TRUE;
  }
  if(IN != stdin) fclose(IN);
  if(err){
    printfinfo("%s: registration was over%s", name, fatal ? " with fatal error" : "");
  } else {
    printfinfo("%s: registration completed successfully", name);
  }
  return err ? 1 : 0;
}


/* processing with finding files in a directory */
int procdir(const char *name, const char *dir, int force, int wmax,
            const char *xtype, const char *xcmd, const char *enc){
  ODEUM *odeum;
  VILLA *mtdb;
  char path[ESTPATHBUFSIZ];
  int err, fatal;
  printfinfo("%s: registration started", name);
  if(!(odeum = odopen(name, OD_OWRITER | OD_OCREAT))){
    pdperror(name);
    return 1;
  }
  sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTMTDBNAME);
  if(!(mtdb = vlopen(path, VL_OWRITER | VL_OCREAT | VL_OZCOMP, VL_CMPLEX))){
    pdperror(name);
    odclose(odeum);
    return 1;
  }
  vlsettuning(mtdb, ESTMTDBLRM, ESTMTDBNIM, ESTMTDBLCN, ESTMTDBNCN);
  printfinfo("%s: database opened: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  err = FALSE;
  if(!indexdir(odeum, mtdb, name, dir, force, wmax, xtype, xcmd, enc)) err = TRUE;
  fatal = odfatalerror(odeum);
  printfinfo("%s: database closing: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!vlclose(mtdb)){
    pdperror(name);
    err = TRUE;
  }
  if(!odclose(odeum)){
    pdperror(name);
    err = TRUE;
  }
  if(err){
    printfinfo("%s: registration was over%s", name, fatal ? " with fatal error" : "");
  } else {
    printfinfo("%s: registration completed successfully", name);
  }
  return err ? 1 : 0;
}


/* find and index files in a directory */
int indexdir(ODEUM *odeum, VILLA *mtdb, const char *name, const char *dir, int force, int wmax,
             const char *xtype, const char *xcmd, const char *enc){
  CBLIST *files;
  const char *file;
  char path[ESTPATHBUFSIZ];
  int i, isroot, isdir, err;
  if(!(files = cbdirlist(dir))){
    printfinfo("%s: directory cannot be opened", dir);
    return TRUE;
  }
  isroot = dir[0] == ESTPATHCHR && dir[1] == '\0';
  err = FALSE;
  for(i = 0; i < cblistnum(files); i++){
    if(sigterm){
      printferror("aborting due to a termination signal");
      cblistclose(files);
      return FALSE;
    }
    file = cblistval(files, i, NULL);
    if(!strcmp(file, ESTCDIRSTR) || !strcmp(file, ESTPDIRSTR)) continue;
    if(isroot){
      sprintf(path, "%s%s", dir, file);
    } else {
      sprintf(path, "%s%c%s", dir, ESTPATHCHR, file);
    }
    if(!cbfilestat(path, &isdir, NULL, NULL)){
      printfinfo("%s: file does not exist", file);
      continue;
    }
    if(isdir){
      if(!indexdir(odeum, mtdb, name, path, force, wmax, xtype, xcmd, enc)) err = TRUE;
    } else {
      if(!indexfile(odeum, mtdb, name, path, force, wmax, xtype, xcmd, enc)) err = TRUE;
    }
  }
  cblistclose(files);
  return err ? FALSE : TRUE;
}


/* index a file into the database */
int indexfile(ODEUM *odeum, VILLA *mtdb, const char *name, const char *file, int force, int wmax,
              const char *xtype, const char *xcmd, const char *enc){
  static int cnt = 0;
  char *buf, *uri;
  const char *date, *title;
  int size, mtime, wnum, bnum, err;
  ODDOC *doc;
  if(!cbfilestat(file, NULL, &size, &mtime)){
    printfinfo("%s: file does not exist", file);
    return TRUE;
  }
  date = datestr(mtime);
  doc = NULL;
  if(bwimatchlist(file, tsuflist)){
    uri = filetouri(file);
    if(!force && !ishot(mtdb, file, mtime)){
      printfinfo("%s: passed", uri);
      free(uri);
      return TRUE;
    }
    if(!(buf = cbreadfile(file, &size))){
      printfinfo("%s: file cannot be opened", uri);
      free(uri);
      return TRUE;
    }
    doc = estdocplain(uri, buf, size, enc);
    if(!oddocgetattr(doc, "date")) oddocaddattr(doc, "date", date);
    free(uri);
    free(buf);
  } else if(bwimatchlist(file, hsuflist)){
    uri = filetouri(file);
    if(!force && !ishot(mtdb, file, mtime)){
      printfinfo("%s: passed", uri);
      free(uri);
      return TRUE;
    }
    if(!(buf = cbreadfile(file, NULL))){
      printfinfo("%s: file cannot be opened", uri);
      free(uri);
      return TRUE;
    }
    doc = estdochtml(uri, buf, size, enc);
    if(!oddocgetattr(doc, "date")) oddocaddattr(doc, "date", date);
    free(uri);
    free(buf);
  } else if(bwimatchlist(file, msuflist)){
    uri = filetouri(file);
    if(!force && !ishot(mtdb, file, mtime)){
      printfinfo("%s: passed", uri);
      free(uri);
      return TRUE;
    }
    if(!(buf = cbreadfile(file, NULL))){
      printfinfo("%s: file cannot be opened", file);
      free(uri);
      return TRUE;
    }
    doc = estdocmime(uri, buf, enc);
    if(!oddocgetattr(doc, "date")) oddocaddattr(doc, "date", date);
    free(uri);
    free(buf);
  } else if(xcmd && xtype && bwimatchlist(file, xsuflist)){
    uri = filetouri(file);
    if(!force && !ishot(mtdb, file, mtime)){
      printfinfo("%s: passed", uri);
      free(uri);
      return TRUE;
    }
    if(!(buf = cbreadfile(file, NULL))){
      printfinfo("%s: file cannot be opened", uri);
      free(uri);
      return TRUE;
    }
    doc = docext(uri, buf, size, enc, xtype, xcmd);
    if(!oddocgetattr(doc, "date")) oddocaddattr(doc, "date", date);
    free(uri);
    free(buf);
  }
  err = FALSE;
  if(doc){
    docltitle(doc, file);
    if(!(title = oddocgetattr(doc, "title")) || strlen(title) < 1){
      if((title = strrchr(file, ESTPATHCHR)) != NULL){
        title++;
      }  else {
        title = file;
      }
      if(!strcmp(cbencname(title, -1), "US-ASCII")) oddocaddattr(doc, "title", title);
    }
    if(odput(odeum, doc, wmax, TRUE) &&
       vlput(mtdb, file, -1, (char *)&mtime, sizeof(int), VL_DOVER)){
      printfinfo("%s: registered: id=%d wnum=%d",
                 oddocuri(doc), oddocid(doc), cblistnum(oddocnwords(doc)));
      cnt++;
    } else {
      pdperror(file);
      err = TRUE;
    }
    oddocclose(doc);
  }
  wnum = odwnum(odeum);
  bnum = odbnum(odeum);
  if(wnum != -1 && bnum != -1 && (double)wnum / (double)bnum > ESTMAXLOAD){
    printfinfo("%s: optimizing started: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
               name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
    if(odoptimize(odeum)){
      printfinfo("%s: optimizing completed: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
                 name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
    } else {
      pdperror(file);
      err = TRUE;
    }
  }
  if(cnt >= 256){
    printfinfo("%s: database status: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
               name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
    cnt = 0;
  }
  return err ? FALSE : TRUE;
}


/* check whetehr a document of the path is new or modified. */
int ishot(VILLA *mtdb, const char *path, int mtime){
  char *vbuf;
  int hot, vsiz;
  hot = TRUE;
  if((vbuf = vlget(mtdb, path, -1, &vsiz)) != NULL){
    if(vsiz == sizeof(int) && mtime <= *(int *)vbuf) hot = FALSE;
    free(vbuf);
  }
  return hot;
}


/* register scores of documents */
int procrelate(const char *name, const char *prefix, int force){
  ODEUM *odeum;
  DEPOT *scdb, *dtdb;
  ODDOC *doc;
  VILLA *rdocs;
  CBMAP *scores;
  const char *dstr;
  char path[ESTPATHBUFSIZ], *uri, *vbuf, *mbuf;
  int i, bnum, omode, err, cnt, fatal, vsiz, id, msiz, date;
  printfinfo("%s: relating started", name);
  if(!(odeum = odopen(name, OD_OWRITER))){
    pdperror(name);
    return 1;
  }
  rdocs = odidbrdocs(odeum);
  sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTSCDBNAME);
  bnum = oddnum(odeum) * ESTSCDBBRAT;
  omode = DP_OWRITER | DP_OCREAT;
  if(prefix[0] == '\0' && force) omode |= DP_OTRUNC;
  if(!(scdb = dpopen(path, omode, bnum))){
    pdperror(name);
    odclose(odeum);
    return 1;
  }
  sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTDTDBNAME);
  bnum = oddnum(odeum) * ESTDTDBBRAT;
  omode = DP_OWRITER | DP_OCREAT;
  if(prefix[0] == '\0' && force) omode |= DP_OTRUNC;
  if(!(dtdb = dpopen(path, omode, bnum))){
    pdperror(name);
    dpclose(scdb);
    odclose(odeum);
    return 1;
  }
  printfinfo("%s: database opened: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  err = FALSE;
  if(!estiterinit(odeum, prefix)){
    pdperror(name);
    err = TRUE;
  } else {
    cnt = 0;
    for(i = 0; TRUE; i++){
      if(sigterm){
        printferror("aborting due to a termination signal");
        err = TRUE;
        break;
      }
      if(!(uri = estiternext(odeum, prefix))){
        if(dpecode != DP_ENOITEM){
          pdperror(name);
          err = TRUE;
        }
        break;
      }
      if(!force && (vbuf = vlget(rdocs, uri, -1, &vsiz)) != NULL){
        if(vsiz == sizeof(int) && dpvsiz(scdb, vbuf, vsiz) != -1){
          printfinfo("%s: passed", uri);
          free(vbuf);
          free(uri);
          continue;
        }
        free(vbuf);
      }
      if(!(doc = odget(odeum, uri))){
        if(dpecode != DP_ENOITEM){
          pdperror(name);
          err = TRUE;
        }
        free(uri);
        break;
      }
      id = oddocid(doc);
      scores = oddocscores(doc, ESTKEYNUM, odeum);
      mbuf = cbmapdump(scores, &msiz);
      dstr = oddocgetattr(doc, "date");
      date = dstr ? eststrmktime(dstr) : -1;
      if(!dpput(scdb, (char *)&id, sizeof(int), mbuf, msiz, DP_DOVER)){
        pdperror(name);
        err = TRUE;
      } else if(!dpput(dtdb, (char *)&id, sizeof(int), (char *)&date, sizeof(int), DP_DOVER)){
        pdperror(name);
        err = TRUE;
      } else {
        printfinfo("%s: related: id=%d", uri, id);
      }
      free(mbuf);
      cbmapclose(scores);
      oddocclose(doc);
      free(uri);
      cnt++;
      if(cnt >= 256){
        printfinfo("%s: progress: %d/%d", name, i + 1, oddnum(odeum));
        cnt = 0;
      }
      if(err) break;
    }
  }
  fatal = odfatalerror(odeum);
  printfinfo("%s: database closing: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!dpclose(dtdb)){
    pdperror(name);
    err = TRUE;
  }
  if(!dpclose(scdb)){
    pdperror(name);
    err = TRUE;
  }
  if(!odclose(odeum)){
    pdperror(name);
    err = TRUE;
  }
  if(err){
    printfinfo("%s: relating was over%s", name, fatal ? " with fatal error" : "");
  } else {
    printfinfo("%s: relating completed successfully", name);
  }
  return err ? 1 : 0;
}


/* purge documents which did not exist. */
int procpurge(const char *name, const char *prefix, int force){
  ODEUM *odeum;
  char *uri, *file;
  int err, fatal;
  printfinfo("%s: purging started", name);
  if(!(odeum = odopen(name, OD_OWRITER))){
    pdperror(name);
    return 1;
  }
  printfinfo("%s: database opened: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  err = FALSE;
  if(!estiterinit(odeum, prefix)){
    pdperror(name);
    err = TRUE;
  } else {
    while(TRUE){
      if(sigterm){
        printferror("aborting due to a termination signal");
        err = TRUE;
        break;
      }
      if(!(uri = estiternext(odeum, prefix))){
        if(dpecode != DP_ENOITEM){
          pdperror(name);
          err = TRUE;
        }
        break;
      }
      file = uritofile(uri);
      if(!force && cbfilestat(file, NULL, NULL, NULL)){
        printfinfo("%s: passed", uri);
      } else {
        if(!odout(odeum, uri)){
          pdperror(file);
          err = TRUE;
        }
        if(!estiterresurge(odeum, uri)){
          pdperror(file);
          err = TRUE;
        }
        printfinfo("%s: purged", uri);
      }
      free(file);
      free(uri);
    }
  }
  fatal = odfatalerror(odeum);
  printfinfo("%s: database closing: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!odclose(odeum)){
    pdperror(name);
    err = TRUE;
  }
  if(err){
    printfinfo("%s: purging was over%s", name, fatal ? " with fatal error" : "");
  } else {
    printfinfo("%s: purging completed successfully", name);
  }
  return err ? 1 : 0;
}


/* optimize a database */
int procoptimize(const char *name){
  ODEUM *odeum;
  if(!(odeum = odopen(name, OD_OWRITER))){
    pdperror(name);
    return 1;
  }
  printfinfo("%s: optimizing started: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!odoptimize(odeum)){
    pdperror(name);
    odclose(odeum);
    return 1;
  }
  printfinfo("%s: optimizing completed: fsiz=%.0f dnum=%d wnum=%d bnum=%d",
             name, odfsiz(odeum), oddnum(odeum), odwnum(odeum), odbnum(odeum));
  if(!odclose(odeum)){
    pdperror(name);
    return 1;
  }
  return 0;
}


/* print information of a database */
int procinform(const char *name){
  ODEUM *odeum;
  char *tmp;
  if(!(odeum = odopen(name, OD_OREADER))){
    pdperror(name);
    return 1;
  }
  tmp = odname(odeum);
  printf("name: %s\n", tmp ? tmp : "(null)");
  free(tmp);
  printf("file size: %.0f\n", odfsiz(odeum));
  printf("index buckets: %d\n", odbnum(odeum));
  printf("used buckets: %d\n", odbusenum(odeum));
  printf("all documents: %d\n", oddnum(odeum));
  printf("all words: %d\n", odwnum(odeum));
  printf("inode number: %d\n", odinode(odeum));
  if(!odclose(odeum)){
    pdperror(name);
    return 1;
  }
  return 0;
}


/* merge some databases */
int procmerge(const char *name, const CBLIST *elems){
  VILLA *mtdb, *emtdb;
  char path[ESTPATHBUFSIZ], *kbuf, *vbuf;
  int i, err, ksiz, vsiz;
  printfinfo("%s: merging started", name);
  if(!odmerge(name, elems)){
    pdperror(name);
    return 1;
  }
  sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTMTDBNAME);
  if(!(mtdb = vlopen(path, VL_OWRITER | VL_OCREAT | VL_OTRUNC | VL_OZCOMP, VL_CMPLEX))){
    pdperror(name);
    return 1;
  }
  err = FALSE;
  for(i = 0; i < cblistnum(elems); i++){
    sprintf(path, "%s%c%s", cblistval(elems, i, NULL), ESTPATHCHR, ESTMTDBNAME);
    if(!(emtdb = vlopen(path, VL_OREADER, VL_CMPLEX))){
      pdperror(name);
      err = TRUE;
      break;
    }
    vlcurfirst(emtdb);
    while(TRUE){
      kbuf = vlcurkey(emtdb, &ksiz);
      vbuf = vlcurval(emtdb, &vsiz);
      if(kbuf && vbuf) vlput(mtdb, kbuf, ksiz, vbuf, vsiz, VL_DKEEP);
      free(vbuf);
      free(kbuf);
      if(!vlcurnext(emtdb)) break;
    }
    if(!vlclose(emtdb)){
      pdperror(name);
      err = TRUE;
      break;
    }
  }
  if(!vlclose(mtdb)){
    pdperror(name);
    err = TRUE;
  }
  if(err){
    printfinfo("%s: merging failed", name);
  } else {
    printfinfo("%s: merging completed successfully", name);
  }
  return err ? 1 : 0;
}


/* print attributes and words extracted from a text */
int procpree(const char *path, int hb, int mb,
             const char *xtype, const char *xcmd, const char *enc){
  ODDOC *doc;
  char *text, *ubuf;
  int size;
  if(path){
    if(!(text = cbreadfile(path, &size))){
      fprintf(stderr, "%s: %s: cannot open\n", progname, path);
      return 1;
    }
  } else {
    path = "STDIN";
    text = readstdin(&size);
  }
  ubuf = filetouri(path);
  if(hb){
    doc = estdochtml(ubuf, text, size, enc);
  } else if(mb){
    doc = estdocmime(ubuf, text, enc);
  } else if(xtype && xcmd){
    doc = docext(ubuf, text, size, enc, xtype, xcmd);
  } else {
    doc = estdocplain(ubuf, text, size, enc);
  }
  docltitle(doc, path);
  free(ubuf);
  showdoc(doc);
  oddocclose(doc);
  free(text);
  return 0;
}


/* make a document handle using an extension command. */
ODDOC *docext(const char *path, const char *ptr, int size, const char *enc,
              const char *xtype, const char *xcmd){
  ODDOC *doc;
  FILE *OUT;
  ESTFILTER filter;
  char cmd[ESTPATHBUFSIZ], *text, numbuf[ESTNUMBUFSIZ];
  int i, tsiz;
  if((OUT = fopen(EXTINFILE, "wb")) != NULL){
    for(i = 0; i < size; i++){
      fputc(ptr[i], OUT);
    }
    fclose(OUT);
  }
  estputenv("ESTORIG", path);
  if(xcmd[0] == DLFILTERSUF){
    if((filter = estfilterget(xcmd + 1)) != NULL){
      filter(EXTINFILE, EXTOUTFILE);
    } else {
      fprintf(stderr, "%s: no valid function in %s\n", progname, xcmd + 1);
    }
  } else {
    sprintf(cmd, "%s %s %s", xcmd, EXTINFILE, EXTOUTFILE);
    system(cmd);
  }
  if((text = cbreadfile(EXTOUTFILE, &tsiz)) != NULL){
    if(xtext){
      doc = estdocplain(path, text, tsiz, enc);
    } else if(xmime){
      doc = estdocmime(path, text, enc);
    } else {
      doc = estdochtml(path, text, tsiz, enc);
    }
    free(text);
  } else {
    doc = oddocopen(path);
  }
  oddocaddattr(doc, "type", xtype);
  sprintf(numbuf, "%d", size);
  oddocaddattr(doc, "size", numbuf);
  remove(EXTOUTFILE);
  remove(EXTINFILE);
  return doc;
}


/* set the local path of a file as its title */
void docltitle(ODDOC *doc, const char *path){
  char *sel, *title;
  if(ptcode || ftcode){
    if(cbstrfwmatch(path, ESTCDIRSTR) && path[sizeof(ESTCDIRSTR)-1] == ESTPATHCHR)
      path += sizeof(ESTCDIRSTR);
    if((sel = cbiconv(path, -1, ptcode ? ptcode : ftcode, "UTF-8", NULL, NULL)) != NULL){
      if(ftcode && (title = strrchr(sel, ESTPATHCHR)) != NULL){
        title++;
      } else {
        title = sel;
      }
      oddocaddattr(doc, "title", title);
      estdocaddtext(doc, title, strlen(title), "UTF-8", ESTDOCNONLY);
      free(sel);
    }
  }
}


/* show attributes and words of a document */
void showdoc(ODDOC *doc){
  const char *tmp;
  const CBLIST *nwords, *awords;
  int i;
  printf("URI: %s\n", oddocuri(doc));
  if((tmp = oddocgetattr(doc, "title")) != NULL) printf("Title: %s\n", tmp);
  if((tmp = oddocgetattr(doc, "author")) != NULL) printf("Author: %s\n", tmp);
  if((tmp = oddocgetattr(doc, "date")) != NULL) printf("Date: %s\n", tmp);
  if((tmp = oddocgetattr(doc, "type")) != NULL) printf("Type: %s\n", tmp);
  if((tmp = oddocgetattr(doc, "encoding")) != NULL) printf("Encoding: %s\n", tmp);
  if((tmp = oddocgetattr(doc, "size")) != NULL) printf("Size: %s\n", tmp);
  printf("\n");
  nwords = oddocnwords(doc);
  awords = oddocawords(doc);
  for(i = 0; i < cblistnum(nwords); i++){
    printf("%s\t%s\n", cblistval(nwords, i, NULL), cblistval(awords, i, NULL));
  }
}



/* END OF FILE */
