/*************************************************************************************************
 * Command line utility to view documents in index
 *                                                      Copyright (C) 2003-2004 Mikio Hirabayashi
 * This file is part of Estraier, a personal full-text search system.
 * Estraier is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 * Estraier is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with Estraier;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA.
 *************************************************************************************************/


#include "estcommon.h"

#define CSSFILE     "estxview.css"       /* name of the CSS file */
#define XSLFILE     "estxview.xsl"       /* name of the XSL file */
#define DTDFILE     "estxview.dtd"       /* name of the DTD file */
#define DEFMAX      8                    /* number of shown documents by default */
#define RELKEYS     16                   /* number of words used with relational search */
#define RELVECNUM   32                   /* number of dimension of score vector */
#define RESROOM     8                    /* ratio of hits for max */
#define UNITINIT    1024                 /* unit number of boolean search */
#define UNITINC     4                    /* inclement coefficient of boolean unit */
#define REEVMAX     1024                 /* max number of candidates of a regular expression */
#define DEFSNUM     80                   /* number of words in summary by default */
#define SUMTOP      16                   /* number of words at the top of summary */
#define SUMWIDTH    16                   /* number of words near each keyword in summary */


/* global variables */
const char *progname = NULL;             /* program name */
ODEUM *odeum = NULL;                     /* handle of the index */
DEPOT *scdb = NULL;                      /* handle of the score database */
DEPOT *dtdb = NULL;                      /* handle of the date database */
int opttiny = FALSE;                     /* whether to show documents in tiny format */
int optnt = FALSE;                       /* whether to hide text of documents */
int optsnum = DEFSNUM;                   /* number of words in summary */
int optnk = FALSE;                       /* whether to hide keywords of documents */
int optcss = FALSE;                      /* whether to show the process instruction for CSS */
int optxsl = FALSE;                      /* whether to show the process instruction for XSLT */
int optdtd = FALSE;                      /* whether to show the DTD reference */


/* function prototypes */
int main(int argc, char **argv);
void usage(void);
int dosearch(const char *name, const char *expr, int ib, int ub, int rel, int relsc,
             const char *etype, const char *stype, int max, int ni);
void xmlprintf(const char *format, ...);
void showerror(const char *msg, int code);
const char *getoptionstr(void);
void showdoc(const ODDOC *doc, int score, double rel, ESTWORD *words, int wnum);
void showsummary(const ODDOC *doc, const ESTWORD *words, int wnum);
void showdoctiny(int id, int score, double rel);
int showdocbyid(int id);
int showdocbyuri(const char *uri);
int showdocbyrel(int id, const char *relsc, int max, int ni);
int showdocbywords(ESTWORD *words, int wnum, const char *etype, const char *stype,
                   int max, int ni);
void setovec(CBMAP *scores, int *vec);
void settvec(CBMAP *osc, CBMAP *tsc, int *vec);


/* main routine */
int main(int argc, char **argv){
  CBDATUM *expr;
  char *name, *etype, *stype, *tmp;
  int i, ib, ub, rel, relsc, max, ni, rf;
  estputenv("LANG", ESTLOCALE);
  estputenv("LC_ALL", ESTLOCALE);
  cbstdiobin();
  progname = argv[0];
  name = NULL;
  expr = cbdatumopen("", 0);
  cbglobalgc(expr, (void (*)(void *))cbdatumclose);
  ib = FALSE;
  ub = FALSE;
  rel = FALSE;
  relsc = FALSE;
  etype = "asis";
  stype = "score";
  max = 0;
  ni = FALSE;
  rf = FALSE;
  for(i = 1; i < argc; i++){
    if(!name && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-id")){
        ib = TRUE;
      } else if(!strcmp(argv[i], "-uri")){
        ub = TRUE;
      } else if(!strcmp(argv[i], "-rel")){
        rel = TRUE;
      } else if(!strcmp(argv[i], "-relsc")){
        relsc = TRUE;
      } else if(!strcmp(argv[i], "-expr")){
        if(++i >= argc) usage();
        etype = argv[i];
      } else if(!strcmp(argv[i], "-sort")){
        if(++i >= argc) usage();
        stype = argv[i];
      } else if(!strcmp(argv[i], "-max")){
        if(++i >= argc) usage();
        max = atoi(argv[i]);
      } else if(!strcmp(argv[i], "-ni")){
        ni = TRUE;
      } else if(!strcmp(argv[i], "-tiny")){
        opttiny = TRUE;
      } else if(!strcmp(argv[i], "-nt")){
        optnt = TRUE;
      } else if(!strcmp(argv[i], "-snum")){
        if(++i >= argc) usage();
        optsnum = atoi(argv[i]);
      } else if(!strcmp(argv[i], "-nk")){
        optnk = TRUE;
      } else if(!strcmp(argv[i], "-css")){
        optcss = TRUE;
      } else if(!strcmp(argv[i], "-xsl")){
        optxsl = TRUE;
      } else if(!strcmp(argv[i], "-dtd")){
        optdtd = TRUE;
      } else if(!strcmp(argv[i], "-rf")){
        rf = TRUE;
      } else {
        usage();
      }
    } else if(!name){
      name = argv[i];
    } else {
      if(cbdatumsize(expr) > 0) cbdatumcat(expr, " ", -1);
      cbdatumcat(expr, argv[i], -1);
    }
  }
  if(!name) usage();
  if(strcmp(etype, "asis") && strcmp(etype, "wild") && strcmp(etype, "regex")) usage();
  if(strcmp(stype, "score") && strcmp(stype, "r-score") &&
     strcmp(stype, "date") && strcmp(stype, "r-date")) usage();
  if(max < 1) max = DEFMAX;
  if(rf && cbdatumsize(expr) > 0){
    if(!(tmp = cbreadfile(cbdatumptr(expr), NULL))){
      fprintf(stderr, "%s: %s: can not open\n", progname, cbdatumptr(expr));
      exit(1);
    }
    cbdatumsetsize(expr, 0);
    cbdatumcat(expr, tmp, -1);
    free(tmp);
  }
  if(cbstrfwimatch(cbdatumptr(expr), "[related] ")){
    rel = TRUE;
    tmp = cbmemdup(strchr(cbdatumptr(expr), ' ') + 1, -1);
    cbdatumsetsize(expr, 0);
    cbdatumcat(expr, tmp, -1);
    free(tmp);
  } else if(cbstrfwimatch(cbdatumptr(expr), "[detail] ")){
    ib = TRUE;
    tmp = cbmemdup(strchr(cbdatumptr(expr), ' ') + 1, -1);
    cbdatumsetsize(expr, 0);
    cbdatumcat(expr, tmp, -1);
    free(tmp);
  }
  return dosearch(name, cbdatumptr(expr), ib, ub, rel, relsc, etype, stype, max, ni);
}


/* print the usage and exit */
void usage(void){
  fprintf(stderr, "%s: viewer documents in index\n", progname);
  fprintf(stderr, "\n");
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "  %s [-id] [-uri] [-rel] [-relsc] [-expr type] [-sort type] [-max num]"
          " [-ni] [-tiny] [-nt] [-snum num] [-nk] [-css] [-xsl] [-dtd] [-rf]"
          " name expression...\n", progname);
  fprintf(stderr, "\n");
  exit(1);
}


/* search documents */
int dosearch(const char *name, const char *expr, int ib, int ub, int rel, int relsc,
             const char *etype, const char *stype, int max, int ni){
  ESTWORD *words;
  char path[ESTPATHBUFSIZ];
  int err, id, wnum;
  err = FALSE;
  if(!(odeum = odopen(name, OD_OREADER))){
    fprintf(stderr, "%s: %s: could not open: %s\n", progname, name, dperrmsg(dpecode));
    err = TRUE;
  } else {
    sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTSCDBNAME);
    scdb = dpopen(path, DP_OREADER, -1);
    sprintf(path, "%s%c%s", name, ESTPATHCHR, ESTDTDBNAME);
    dtdb = dpopen(path, DP_OREADER, -1);
    if(ib){
      if((id = atoi(expr)) < 1){
        fprintf(stderr, "%s: %s: invalid ID expression\n", progname, expr);
        err = TRUE;
      } else {
        if(!showdocbyid(id)) err = TRUE;
      }
    } else if(ub){
      if(strlen(expr) < 1){
        fprintf(stderr, "%s: %s: invalid URI expression\n", progname, expr);
        err = TRUE;
      } else {
        if(!showdocbyuri(expr)) err = TRUE;
      }
    } else if(rel || relsc){
      if(!scdb){
        fprintf(stderr, "%s: %s: could not open\n", progname, path);
        err = TRUE;
      } else if(relsc){
        if(!showdocbyrel(-1, expr, max, ni)) err = TRUE;
      } else {
        if((id = atoi(expr)) < 1){
          fprintf(stderr, "%s: %s: invalid ID expression\n", progname, expr);
          err = TRUE;
        } else {
          if(!showdocbyrel(id, NULL, max, ni)) err = TRUE;
        }
      }
    } else {
      words = estsearchwords(expr, &wnum, !strcmp(etype, "asis"));
      if(wnum < 1){
        fprintf(stderr, "%s: %s: no effective word in the phrase\n", progname, expr);
        err = TRUE;
      } else {
        if(!showdocbywords(words, wnum, etype, stype, max, ni)) err = TRUE;
      }
      estfreewords(words, wnum);
    }
    if(dtdb) dpclose(dtdb);
    if(scdb) dpclose(scdb);
  }
  if(odeum) odclose(odeum);
  return err ? 1 : 0;
}


/* XML-oriented printf */
void xmlprintf(const char *format, ...){
  va_list ap;
  char *tmp;
  unsigned char c;
  va_start(ap, format);
  while(*format != '\0'){
    if(*format == '%'){
      format++;
      switch(*format){
      case 's':
        tmp = va_arg(ap, char *);
        if(!tmp) tmp = "(null)";
        printf("%s", tmp);
        break;
      case 'd':
        printf("%d", va_arg(ap, int));
        break;
      case '@':
        tmp = va_arg(ap, char *);
        if(!tmp) tmp = "(null)";
        while(*tmp){
          switch(*tmp){
          case '&': printf("&amp;"); break;
          case '<': printf("&lt;"); break;
          case '>': printf("&gt;"); break;
          case '"': printf("&quot;"); break;
          default: putchar(*tmp); break;
          }
          tmp++;
        }
        break;
      case '?':
        tmp = va_arg(ap, char *);
        if(!tmp) tmp = "(null)";
        while(*tmp){
          c = *(unsigned char *)tmp;
          if((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') ||
             (c >= '0' && c <= '9') || (c != '\0' && strchr("_-.", c))){
            putchar(c);
          } else {
            printf("%%%02X", c);
          }
          tmp++;
        }
        break;
      case '%':
        putchar('%');
        break;
      }
    } else {
      putchar(*format);
    }
    format++;
  }
  va_end(ap);
}


/* show an error message */
void showerror(const char *msg, int code){
  if(code >= 0){
    xmlprintf("<error code=\"%d\" string=\"%@\">%@</error>\n", code, dperrmsg(dpecode), msg);
  } else {
    xmlprintf("<error>%@</error>\n", msg);
  }
}


/* get the string of the building options */
const char *getoptionstr(void){
  static char str[ESTPATHBUFSIZ], *wp;
  wp = str;
  if(estisregex){
    if(wp != str) *(wp++) = ' ';
    wp += sprintf(wp, "regex");
  }
  if(ESTISSTRICT){
    if(wp != str) *(wp++) = ' ';
    wp += sprintf(wp, "strict");
  }
  if(ESTISNOSTOPW){
    if(wp != str) *(wp++) = ' ';
    wp += sprintf(wp, "no-stopword");
  }
  if(estischasen){
    if(wp != str) *(wp++) = ' ';
    wp += sprintf(wp, "chasen");
  }
  if(estiskakasi){
    if(wp != str) *(wp++) = ' ';
    wp += sprintf(wp, "kakasi");
  }
  *wp = '\0';
  return str;
}


/* set the original score vector */
void setovec(CBMAP *scores, int *vec){
  int i;
  const char *kbuf;
  cbmapiterinit(scores);
  for(i = 0; i < RELVECNUM; i++){
    if((kbuf = cbmapiternext(scores, NULL)) != NULL){
      vec[i] = atoi(cbmapget(scores, kbuf, -1, NULL));
    } else {
      vec[i] = 0;
    }
  }
}


/* set the target score vector */
void settvec(CBMAP *osc, CBMAP *tsc, int *vec){
  int i;
  const char *kbuf, *vbuf;
  cbmapiterinit(osc);
  for(i = 0; i < RELVECNUM; i++){
    if((kbuf = cbmapiternext(osc, NULL)) != NULL){
      vbuf = cbmapget(tsc, kbuf, -1, NULL);
      vec[i] = vbuf ? atoi(vbuf) : 0;
    } else {
      vec[i] = 0;
    }
  }
}


/* show a document */
void showdoc(const ODDOC *doc, int score, double rel, ESTWORD *words, int wnum){
  const CBLIST *nwords, *awords;
  CBMAP *scores;
  CBDATUM *ssel;
  const char *uri, *title, *author, *rcpt, *mcast, *date, *type, *enc, *size, *word;
  char numbuf[ESTNUMBUFSIZ], *mbuf;
  int i, id, msiz, wsiz;
  id = oddocid(doc);
  uri = oddocuri(doc);
  title = oddocgetattr(doc, "title");
  author = oddocgetattr(doc, "author");
  rcpt = oddocgetattr(doc, "recipient");
  mcast = oddocgetattr(doc, "multicast");
  date = oddocgetattr(doc, "date");
  type = oddocgetattr(doc, "type");
  enc = oddocgetattr(doc, "encoding");
  size = oddocgetattr(doc, "size");
  if(rel >= 0.0){
    sprintf(numbuf, "%.2f", rel / 100);
    xmlprintf("<document id=\"%d\" similarity=\"%@\">\n", id, numbuf);
  } else if(score >= 0){
    xmlprintf("<document id=\"%d\" score=\"%d\">\n", id, score);
  } else {
    xmlprintf("<document id=\"%d\">\n", id);
  }
  xmlprintf("<uri>%@</uri>\n", uri);
  if(title && title[0] != '\0') xmlprintf("<title>%@</title>\n", title);
  if(author && author[0] != '\0') xmlprintf("<author>%@</author>\n", author);
  if(rcpt && rcpt[0] != '\0') xmlprintf("<recipient>%@</recipient>\n", rcpt);
  if(mcast && mcast[0] != '\0') xmlprintf("<multicast>%@</multicast>\n", mcast);
  if(date && date[0] != '\0') xmlprintf("<date>%@</date>\n", date);
  if(type && type[0] != '\0') xmlprintf("<type>%@</type>\n", type);
  if(enc && enc[0] != '\0') xmlprintf("<encoding>%@</encoding>\n", enc);
  if(size && size[0] != '\0') xmlprintf("<size>%@</size>\n", size);
  if(!optnt){
    nwords = oddocnwords(doc);
    awords = oddocawords(doc);
    xmlprintf("<text>\n");
    for(i = 0; i < cblistnum(nwords); i++){
      xmlprintf("<word normal=\"%@\">%@</word>\n",
                cblistval(nwords, i, NULL), cblistval(awords, i, NULL));
    }
    xmlprintf("</text>\n");
  }
  if(optsnum > 0) showsummary(doc, words, wnum);
  if(!optnk && scdb && (mbuf = dpget(scdb, (char *)&id, sizeof(int), 0, -1, &msiz)) != NULL){
    ssel = cbdatumopen("", 0);
    scores = cbmapload(mbuf, msiz);
    cbmapiterinit(scores);
    while((word = cbmapiternext(scores, &wsiz)) != NULL){
      cbdatumcat(ssel, word, wsiz);
      cbdatumcat(ssel, "\t", 1);
      cbdatumcat(ssel, cbmapget(scores, word, wsiz, NULL), -1);
      cbdatumcat(ssel, "\n", 1);
    }
    xmlprintf("<keywords>%?</keywords>\n", cbdatumptr(ssel));
    cbdatumclose(ssel);
    cbmapclose(scores);
    free(mbuf);
  }
  xmlprintf("</document>\n");
}


/* show summary of a document */
void showsummary(const ODDOC *doc, const ESTWORD *words, int wnum){
  const CBLIST *nwords, *awords;
  CBMAP *kmap, *tmap;
  const char *evword, *normal, *asis, *ibuf;
  int i, j, lnum, gshow, lshow, nwsiz, awsiz, cjk, tb, pv, bi;
  xmlprintf("<summary>");
  nwords = oddocnwords(doc);
  awords = oddocawords(doc);
  kmap = cbmapopen();
  tmap = cbmapopen();
  if(words){
    for(i = 0; i < wnum; i++){
      if(words[i].type == ESTCONDNOT) continue;
      if(words[i].evwords){
        for(j = 0; j < cblistnum(words[i].evwords); j++){
          evword = cblistval(words[i].evwords, j, NULL);
          cbmapput(kmap, evword, -1, (char *)&i, sizeof(int), TRUE);
          cbmapput(tmap, evword, -1, (char *)&i, sizeof(int), TRUE);
        }
      } else {
        cbmapput(kmap, words[i].word, -1, (char *)&i, sizeof(int), TRUE);
        cbmapput(tmap, words[i].word, -1, (char *)&i, sizeof(int), TRUE);
      }
    }
  }
  lnum = cblistnum(nwords);
  gshow = 0;
  /* show top words */
  lshow = 0;
  cjk = FALSE;
  for(i = 0; i < lnum && lshow < (words ? SUMTOP : optsnum); i++){
    normal = cblistval(nwords, i, &nwsiz);
    asis = cblistval(awords, i, &awsiz);
    if(awsiz < 1) continue;
    if(lshow > 0 && (!cjk || *(unsigned char *)asis < 0xe0)) xmlprintf(" ");
    if(nwsiz > 0 && (ibuf = cbmapget(kmap, normal, nwsiz, NULL)) != NULL){
      xmlprintf("<key>%@</key>", asis);
    } else {
      xmlprintf("%@", asis);
    }
    cjk = *(unsigned char *)asis >= 0xe0;
    gshow++;
    lshow++;
  }
  xmlprintf(" <delim>...</delim> ");
  /* show phrases around search words */
  tb = wnum > 0 && wnum < 4 ? (SUMWIDTH / 2) / wnum : 0;
  pv = i;
  while(i < lnum){
    if(cbmaprnum(kmap) > 0 && cbmaprnum(tmap) < 1){
      cbmapclose(tmap);
      tmap = cbmapdup(kmap);
    }
    normal = cblistval(nwords, i, &nwsiz);
    if(cbmapget(tmap, normal, nwsiz, NULL)){
      cbmapout(tmap, normal, nwsiz);
      bi = i - (SUMWIDTH + tb) / 2;
      bi = bi > pv ? bi : pv;
      lshow = 0;
      cjk = FALSE;
      for(j = bi; j < lnum && j <= bi + SUMWIDTH + tb; j++){
        normal = cblistval(nwords, j, &nwsiz);
        asis = cblistval(awords, j, &awsiz);
        if(awsiz < 1) continue;
        if(lshow > 0 && (!cjk || *(unsigned char *)asis < 0xe0)) xmlprintf(" ");
        if(nwsiz > 0 && (ibuf = cbmapget(kmap, normal, nwsiz, NULL)) != NULL){
          xmlprintf("<key>%@</key>", asis);
        } else {
          xmlprintf("%@", asis);
        }
        cjk = *(unsigned char *)asis >= 0xe0;
        gshow++;
        lshow++;
      }
      xmlprintf(" <delim>...</delim> ");
      i = j;
      pv = i;
    } else {
      i++;
    }
    if(gshow > optsnum) break;
  }
  /* infill the left space */
  if(pv < lnum - SUMWIDTH && optsnum - gshow > SUMWIDTH){
    lshow = 0;
    cjk = FALSE;
    for(i = pv; i < lnum && gshow <= optsnum; i++){
      asis = cblistval(awords, i, NULL);
      if(lshow > 0 && (!cjk || *(unsigned char *)asis < 0xe0)) xmlprintf(" ");
      xmlprintf("%@", asis);
      cjk = *(unsigned char *)asis >= 0xe0;
      gshow++;
      lshow++;
    }
    xmlprintf(" <delim>...</delim> ");
  }
  /* release resources */
  cbmapclose(tmap);
  cbmapclose(kmap);
  xmlprintf("</summary>\n");
}


/* show a document in tiny format */
void showdoctiny(int id, int score, double rel){
  char numbuf[ESTNUMBUFSIZ];
  if(rel >= 0.0){
    sprintf(numbuf, "%.2f", rel / 100);
    xmlprintf("<document id=\"%d\" similarity=\"%@\"/>\n", id, numbuf);
  } else if(score >= 0){
    xmlprintf("<document id=\"%d\" score=\"%d\"/>\n", id, score);
  } else {
    xmlprintf("<document id=\"%d\"/>\n", id);
  }
}


/* show a document specified by the ID */
int showdocbyid(int id){
  ODDOC *doc;
  int err, hit;
  err = FALSE;
  xmlprintf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  if(optcss) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"text/css\"?>\n", CSSFILE);
  if(optxsl) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"application/xml\"?>\n", XSLFILE);
  if(optdtd) xmlprintf("<!DOCTYPE estxview SYSTEM \"%@\">\n", DTDFILE);
  xmlprintf("<estxview version=\"%@\" options=\"%@\">\n", _EST_VERSION, getoptionstr());
  xmlprintf("<meta>\n");
  xmlprintf("<query operator=\"id\">%d</query>\n", id);
  doc = NULL;
  hit = 0;
  if(opttiny){
    if(odcheck(odeum, id)){
      hit = 1;
    } else {
      id = -1;
      showerror("There is no corresponding document.", dpecode);
      err = TRUE;
    }
  } else {
    if((doc = odgetbyid(odeum, id)) != NULL){
      hit = 1;
    } else {
      showerror("There is no corresponding document.", dpecode);
      err = TRUE;
    }
  }
  xmlprintf("<total hit=\"%d\" dnum=\"%d\" wnum=\"%d\"/>\n", hit, oddnum(odeum), odwnum(odeum));
  xmlprintf("</meta>\n");
  if(opttiny){
    if(id != -1) showdoctiny(id, -1, -1.0);
  } else {
    if(doc){
      showdoc(doc, -1, -1.0, NULL, 0);
      oddocclose(doc);
    }
  }
  xmlprintf("</estxview>\n");
  return err ? FALSE : TRUE;
}


/* show a document specified by the URI */
int showdocbyuri(const char *uri){
  ODDOC *doc;
  int err, hit, id;
  err = FALSE;
  xmlprintf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  if(optcss) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"text/css\"?>\n", CSSFILE);
  if(optxsl) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"application/xml\"?>\n", XSLFILE);
  if(optdtd) xmlprintf("<!DOCTYPE estxview SYSTEM \"%@\">\n", DTDFILE);
  xmlprintf("<estxview version=\"%@\" options=\"%@\">\n", _EST_VERSION, getoptionstr());
  xmlprintf("<meta>\n");
  xmlprintf("<query operator=\"uri\">%@</query>\n", uri);
  hit = 0;
  doc = NULL;
  id = -1;
  if(opttiny){
    if((id = odgetidbyuri(odeum, uri)) != -1){
      hit = 1;
    } else {
      showerror("There is no corresponding document.", dpecode);
      err = TRUE;
    }
  } else {
    if((doc = odget(odeum, uri)) != NULL){
      hit = 1;
    } else {
      showerror("There is no corresponding document.", dpecode);
      err = TRUE;
    }
  }
  xmlprintf("<total hit=\"%d\" dnum=\"%d\" wnum=\"%d\"/>\n", hit, oddnum(odeum), odwnum(odeum));
  xmlprintf("</meta>\n");
  if(opttiny){
    if(id != -1) showdoctiny(id, -1, -1.0);
  } else {
    if(doc){
      showdoc(doc, -1, -1.0, NULL, 0);
      oddocclose(doc);
    }
  }
  xmlprintf("</estxview>\n");
  return err ? FALSE : TRUE;
}


/* show documents related to a document of the ID */
int showdocbyrel(int id, const char *relsc, int max, int ni){
  ODDOC *doc, *tdoc;
  CBMAP *scores, *tsc;
  CBLIST *twords;
  CBDATUM *phrase;
  ESTWORD *words;
  ODPAIR *pairs;
  const char *word, *score;
  char *mbuf, *tmbuf;
  int i, err, msiz, tmsiz, wsiz, wnum, pnum, lnum, show;
  int ovec[RELVECNUM], tvec[RELVECNUM];
  err = FALSE;
  xmlprintf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  if(optcss) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"text/css\"?>\n", CSSFILE);
  if(optxsl) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"application/xml\"?>\n", XSLFILE);
  if(optdtd) xmlprintf("<!DOCTYPE estxview SYSTEM \"%@\">\n", DTDFILE);
  xmlprintf("<estxview version=\"%@\" options=\"%@\">\n", _EST_VERSION, getoptionstr());
  xmlprintf("<meta>\n");
  if(relsc){
    xmlprintf("<query operator=\"relsc\">%?</query>\n", relsc);
  } else {
    xmlprintf("<query operator=\"related\">%d</query>\n", id);
  }
  xmlprintf("<option max=\"%d\" tfidf=\"%@\" expr=\"none\" sort=\"none\"/>\n",
            max, ni ? "false" : "true");
  doc = NULL;
  mbuf = NULL;
  scores = NULL;
  twords = NULL;
  phrase = NULL;
  words = NULL;
  pairs = NULL;
  pnum = 0;
  if(!relsc && !(doc = odgetbyid(odeum, id))){
    showerror("There is no seed document.", dpecode);
    err = TRUE;
  } else if(!relsc && !(mbuf = dpget(scdb, (char *)&id, sizeof(int), 0, -1, &msiz))){
    showerror("The document has no score.", -1);
    err = TRUE;
  } else {
    if(relsc){
      scores = cbmapopen();
      twords = cbsplit(relsc, -1, "\n");
      for(i = 0; i < cblistnum(twords); i++){
        word = cblistval(twords, i, NULL);
        if(!(score = strchr(word, '\t'))) continue;
        cbmapput(scores, word, score - word, score + 1, -1, TRUE);
      }
      cblistclose(twords);
      twords = cbmapkeys(scores);
    } else {
      scores = cbmapload(mbuf, msiz);
      twords = cbmapkeys(scores);
    }
    phrase = cbdatumopen("", 0);
    for(i = 0; i < RELKEYS && i < cblistnum(twords); i++){
      word = cblistval(twords, i, &wsiz);
      if(i > 0) cbdatumcat(phrase, " [OR] ", -1);
      cbdatumcat(phrase, word, wsiz);
    }
    words = estsearchwords(cbdatumptr(phrase), &wnum, FALSE);
    pairs = estsearch(odeum, words, wnum, UNITINIT, !ni, &pnum, &lnum, FALSE, FALSE, 0);
    setovec(scores, ovec);
    for(i = 0; i < pnum; i++){
      if((tmbuf = dpget(scdb, (char *)&(pairs[i].id), sizeof(int), 0, -1, &tmsiz)) != NULL){
        tsc = cbmapload(tmbuf, tmsiz);
        free(tmbuf);
      } else {
        tsc = cbmapopen();
      }
      settvec(scores, tsc, tvec);
      cbmapclose(tsc);
      pairs[i].score = odvectorcosine(ovec, tvec, RELVECNUM) * 10000;
      if(pairs[i].score >= 9999) pairs[i].score = 10000;
    }
    odpairssort(pairs, pnum);
  }
  if(pnum < 1 && !err){
    showerror("There is no corresponding document.", DP_ENOITEM);
    err = TRUE;
  }
  xmlprintf("<total hit=\"%d\" dnum=\"%d\" wnum=\"%d\"/>\n", pnum, oddnum(odeum), odwnum(odeum));
  xmlprintf("</meta>\n");
  if(pairs){
    show = 0;
    for(i = 0; i < pnum && show < max; i++){
      if(opttiny){
        if(!odcheck(odeum, pairs[i].id)) continue;
        showdoctiny(pairs[i].id, -1, pairs[i].score);
        show++;
      } else {
        if(!(tdoc = odgetbyid(odeum, pairs[i].id))) continue;
        showdoc(tdoc, -1, pairs[i].score, NULL, 0);
        oddocclose(tdoc);
        show++;
      }
    }
  }
  if(pairs) free(pairs);
  if(words) estfreewords(words, wnum);
  if(phrase) cbdatumclose(phrase);
  if(twords) cblistclose(twords);
  if(scores) cbmapclose(scores);
  if(mbuf) free(mbuf);
  if(doc) oddocclose(doc);
  xmlprintf("</estxview>\n");
  return err ? FALSE : TRUE;
}


/* show documents by searching with words */
int showdocbywords(ESTWORD *words, int wnum, const char *etype, const char *stype,
                   int max, int ni){
  ODDOC *doc;
  ODPAIR *pairs, pswap;
  const char *dstr;
  int i, err, unit, ldnum, pnum, lnum, date, show;
  err = FALSE;
  xmlprintf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  if(optcss) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"text/css\"?>\n", CSSFILE);
  if(optxsl) xmlprintf("<?xml-stylesheet href=\"%@\" type=\"application/xml\"?>\n", XSLFILE);
  if(optdtd) xmlprintf("<!DOCTYPE estxview SYSTEM \"%@\">\n", DTDFILE);
  xmlprintf("<estxview version=\"%@\" options=\"%@\">\n", _EST_VERSION, getoptionstr());
  xmlprintf("<meta>\n");
  pairs = NULL;
  unit = UNITINIT;
  if(strcmp(stype, "score")) unit = INT_MAX;
  do {
    if(pairs){
      free(pairs);
      unit *= UNITINC;
    }
    pairs = estsearch(odeum, words, wnum, unit, !ni, &pnum, &lnum,
                      !strcmp(etype, "regex"), !strcmp(etype, "wild"), REEVMAX);
  } while(pnum < max * RESROOM && wnum > 1 && lnum > 0);
  if(!strcmp(stype, "date") || !strcmp(stype, "r-date")){
    for(i = 0; i < pnum; i++){
      date = 0;
      if(!dtdb){
        if((doc = odgetbyid(odeum, pairs[i].id)) != NULL){
          if((dstr =  oddocgetattr(doc, "date")) != NULL) date = eststrmktime(dstr);
          oddocclose(doc);
        }
      } else {
        dpgetwb(dtdb, (char *)&(pairs[i].id), sizeof(int), 0, sizeof(int), (char *)&date);
      }
      pairs[i].score = date;
    }
    odpairssort(pairs, pnum);
  }
  if(cbstrfwmatch(stype, "r-")){
    for(i = 0; i < pnum / 2; i++){
      pswap = pairs[i];
      pairs[i] = pairs[pnum-i-1];
      pairs[pnum-i-1] = pswap;
    }
  }
  ldnum = 0;
  for(i = 0; i < wnum; i++){
    ldnum = words[i].dnum;
    if(i < 1){
      xmlprintf("<query hit=\"%d\">%@</query>\n", words[i].dnum, words[i].word);
    } else {
      xmlprintf("<query hit=\"%d\" operator=\"%@\">%@</query>\n", words[i].dnum,
                words[i].type == ESTCONDAND ? "and" : words[i].type == ESTCONDOR ? "or" : "not",
                words[i].word);
    }
  }
  xmlprintf("<option max=\"%d\" tfidf=\"%@\" expr=\"%?\" sort=\"%?\"/>\n",
            max, ni ? "false" : "true", etype, stype);
  if(pnum < 1){
    showerror("There is no corresponding document.", dpecode);
    err = TRUE;
  }
  xmlprintf("<total hit=\"%d\" dnum=\"%d\" wnum=\"%d\"/>\n",
            wnum < 2 ? ldnum : pnum, oddnum(odeum), odwnum(odeum));
  xmlprintf("</meta>\n");
  show = 0;
  for(i = 0; i < pnum && show < max; i++){
    if(opttiny){
      if(!odcheck(odeum, pairs[i].id)) continue;
      showdoctiny(pairs[i].id, pairs[i].score, -1.0);
      show++;
    } else {
      if(!(doc = odgetbyid(odeum, pairs[i].id))) continue;
      showdoc(doc, pairs[i].score, -1.0, words, wnum);
      oddocclose(doc);
      show++;
    }
  }
  free(pairs);
  xmlprintf("</estxview>\n");
  return err ? FALSE : TRUE;
}



/* END OF FILE */
